module ahb_decmux
  (
   input wire 	      BOOT_MODE, // 0=normal, 1=boot_loader
   input wire 	      HCLK,
   input wire 	      HRESETn,
   input wire [31:0]  HADDR,

   // Bus element: CODE memory
   input wire [31:0]  HRDATA_CODE,
   input wire 	      HREADYOUT_CODE,
   input wire 	      HRESP_CODE,
   output wire 	      HSEL_CODE,

   // DATA memory
   input wire [31:0]  HRDATA_DATA,
   input wire 	      HREADYOUT_DATA,
   input wire 	      HRESP_DATA,
   output wire 	      HSEL_DATA,

   // Peripheral #0
   input wire [31:0]  HRDATA_P0,
   input wire 	      HREADYOUT_P0,
   input wire 	      HRESP_P0,
   output wire 	      HSEL_P0,

   // Peripheral #1
   input wire [31:0]  HRDATA_P1,
   input wire 	      HREADYOUT_P1,
   input wire 	      HRESP_P1,
   output wire 	      HSEL_P1,

   // Peripheral #2
   input wire [31:0]  HRDATA_P2,
   input wire 	      HREADYOUT_P2,
   input wire 	      HRESP_P2,
   output wire 	      HSEL_P2,

   // APB bus bridge
   input wire [31:0]  HRDATA_APB,
   input wire 	      HREADYOUT_APB,
   input wire 	      HRESP_APB,
   output wire 	      HSEL_APB,

   // This item is used when nothing other
   // selected
   input wire [31:0]  HRDATA_NOMAP,
   input wire 	      HREADYOUT_NOMAP,
   input wire 	      HRESP_NOMAP,
   output wire 	      HSEL_NOMAP,

   // System ROM
   input wire [31:0]  HRDATA_SROM,
   input wire 	      HREADYOUT_SROM,
   input wire 	      HRESP_SROM,
   output wire 	      HSEL_SROM,

   // Test module
   input wire [31:0]  HRDATA_TEST,
   input wire 	      HREADYOUT_TEST,
   input wire 	      HRESP_TEST,
   output wire 	      HSEL_TEST,

   // Output side of the bus
   output wire [31:0] HRDATA,
   output wire 	      HREADYOUT,
   output wire 	      HRESP,
   output wire 	      HSEL_ANY
   );

   reg 		      boot_mode;
   reg [7:0] 	      rhsel;
   reg [4:0] 	      rpsel;

   wire 	      sel_code0;
   wire 	      sel_code1;
   wire 	      hsel_per;

   wire 	      HSEL_BLD;
   wire 	      HSEL_DDR;
   
   assign sel_code0= (HADDR[31:27] == 5'b0000_0);
   assign sel_code1= (HADDR[31:27] == 5'b0000_1);
   assign HSEL_DATA= (HADDR[31:28] == 4'h2);
   assign HSEL_DDR = (HADDR[31:28] == 4'h4);
   assign hsel_per=  (HADDR[31:28] == 4'h5);
   assign HSEL_APB=  (HADDR[31:28] == 4'h6);
   assign HSEL_TEST= (HADDR[31:28] == 4'h7);
   assign HSEL_SROM= (HADDR[31:28] == 4'hf);

   assign HSEL_P0= hsel_per & (HADDR[27:24] == 4'h0);
   assign HSEL_P1= hsel_per & (HADDR[27:24] == 4'h1);
   assign HSEL_P2= hsel_per & (HADDR[27:24] == 4'h2);

   wire hresetn= HRESETn;
   //BUFG res(.I(HRESETn), .O(hresetn));
   
   always @(posedge HCLK or negedge hresetn)
     if (!hresetn)
       boot_mode<= BOOT_MODE;
     else
       boot_mode<= boot_mode;

   assign HSEL_CODE= boot_mode?sel_code1:sel_code0;
   assign HSEL_BLD = boot_mode?sel_code0:sel_code1;
   
   always @(posedge HCLK or negedge HRESETn)
     begin
	if (!HRESETn)
	  begin
	     rhsel<= 0;
	     rpsel<= 0;
	  end
	else if (HREADYOUT)
	  begin
	     rhsel<= {/*  7 */ HSEL_TEST,
		      /*  6 */ HSEL_DDR,
	              /*  5 */ HSEL_BLD,
		      /*  4 */ HSEL_APB,
		      /*  3 */ HSEL_SROM,
		      /*  2 */ hsel_per,
		      /*  1 */ HSEL_DATA,
		      /*  0 */ HSEL_CODE
		      };
	     rpsel<= {/*  3 */ 1'b0,
		      /*  2 */ HSEL_P2,
		      /*  1 */ HSEL_P1,
		      /*  0 */ HSEL_P0
		       };
	  end
     end
   
   assign HSEL_ANY= HSEL_CODE |
		    HSEL_BLD  |
		    HSEL_DDR  |
		    HSEL_DATA |
		    HSEL_SROM |
		    HSEL_APB  |
		    HSEL_TEST |
		    
		    HSEL_P0   |
		    HSEL_P1   |
		    HSEL_P2
		    ;

   assign HRDATA
     =
      rhsel[0]?HRDATA_CODE:
      rhsel[1]?HRDATA_DATA:
      rhsel[3]?HRDATA_SROM:
      rhsel[4]?HRDATA_APB:
      rhsel[5]?/*HRDATA_BLD*/32'b0:
      rhsel[6]?/*HRDATA_DDR*/32'b0:
      rhsel[7]?HRDATA_TEST:
      rpsel[0]?HRDATA_P0:
      rpsel[1]?HRDATA_P1:
      rpsel[2]?HRDATA_P2:
      HRDATA_NOMAP
      ;

   assign HREADYOUT
     =
      rhsel[0]?HREADYOUT_CODE:
      rhsel[1]?HREADYOUT_DATA:
      rhsel[3]?HREADYOUT_SROM:
      rhsel[4]?HREADYOUT_APB:
      rhsel[5]?/*HREADYOUT_BLD*/1'b1:
      rhsel[6]?/*HREADYOUT_DDR*/1'b1:
      rhsel[7]?HREADYOUT_TEST:
      rpsel[0]?HREADYOUT_P0:
      rpsel[1]?HREADYOUT_P1:
      rpsel[2]?HREADYOUT_P2:
      HREADYOUT_NOMAP
      ;
      
   assign HRESP
     =
      rhsel[0]?HRESP_CODE:
      rhsel[1]?HRESP_DATA:
      rhsel[3]?HRESP_SROM:
      rhsel[4]?HRESP_APB:
      rhsel[5]?/*HRESP_BLD*/1'b0:
      rhsel[6]?/*HRESP_DDR*/1'b0:
      rhsel[7]?HRESP_TEST:
      rpsel[0]?HRESP_P0:
      rpsel[1]?HRESP_P1:
      rpsel[2]?HRESP_P2:
      HRESP_NOMAP
      ;

endmodule // ahb_decmux
