module ahb_in2
  #(
    // Parameter for address width
    parameter    ADDRWIDTH =12)
   (
    input wire 		       HCLK, // Clock
    input wire 		       HRESETn, // Reset
    
    input wire [3:0] 	       ECOREVNUM, // Engineering-change-order revision bits
    
    // AHB connection to master
    input wire 		       HSEL,
    input wire [ADDRWIDTH-1:0] HADDR,
    input wire [1:0] 	       HTRANS,
    input wire [2:0] 	       HSIZE,
    input wire 		       HWRITE,
    input wire 		       HREADY,
    input wire [31:0] 	       HWDATA,
    
    output wire 	       HREADYOUT,
    output wire 	       HRESP,
    output wire [31:0] 	       HRDATA,

    input [31:0] 	       pi,
    input [31:0] 	       pj
    );
   
   
   // ----------------------------------------
   // Internal wires declarations
   
   // Register module interface signals
   wire [ADDRWIDTH-1:0]        reg_addr;
   wire 		       reg_read_en;
   wire 		       reg_write_en;
   wire [3:0] 		       reg_byte_strobe;
   wire [31:0] 		       reg_wdata;
   wire [31:0] 		       reg_rdata;
   
   //-----------------------------------------------------------
   // Module logic start
   //----------------------------------------------------------
   
   
   // Interface block to convert AHB transfers to simple read/write
   // controls.
   cmsdk_ahb_eg_slave_interface
     #(.ADDRWIDTH (ADDRWIDTH))
   u_ahb_eg_slave_interface
     (
      .hclk         (HCLK),
      .hresetn      (HRESETn),
      
      // Input slave port: 32 bit data bus interface
      .hsels        (HSEL),
      .haddrs       (HADDR),
      .htranss      (HTRANS),
      .hsizes       (HSIZE),
      .hwrites      (HWRITE),
      .hreadys      (HREADY),
      .hwdatas      (HWDATA),
      
      .hreadyouts   (HREADYOUT),
      .hresps       (HRESP),
      .hrdatas      (HRDATA),
      
      // Register interface
      .addr         (reg_addr),
      .read_en      (reg_read_en),
      .write_en     (reg_write_en),
      .byte_strobe  (reg_byte_strobe),
      .wdata        (reg_wdata),
      .rdata        (reg_rdata)
      );
   
   // Simple data register block with four 32-bit registers
   reg2in
     #(.ADDRWIDTH (ADDRWIDTH))
   u_ahb_eg_slave_reg
     (
      
      .hclk         (HCLK),
      .hresetn      (HRESETn),
      
      // Register interface
      .addr         (reg_addr),
      .read_en      (reg_read_en),
      .write_en     (reg_write_en),
      .byte_strobe  (reg_byte_strobe),
      .wdata        (reg_wdata),
      .ecorevnum    (ECOREVNUM),
      .rdata        (reg_rdata),

      .pi(pi),
      .pj(pj)
      );

   //-----------------------------------------------------------
   //Module logic end
   //----------------------------------------------------------
`ifdef ARM_AHB_ASSERT_ON
   
 `include "std_ovl_defines.h"
   // ------------------------------------------------------------
   // Assertions
   // ------------------------------------------------------------
   
   wire 		       ovl_trans_req = HREADYS & HSELS & HTRANSS[1];
   
   // Check the reg_write_en signal generated
   assert_next
     #(`OVL_ERROR, 1,1,0,
       `OVL_ASSERT,
       "Error! register write signal was not generated! "
       )
   u_ovl_ahb_eg_slave_reg_write
     (.clk         ( HCLK ),
      .reset_n     (HRESETn),
      .start_event ((ovl_trans_req & HWRITES)),
      .test_expr   (reg_write_en == 1'b1)
      );
   
   
   // Check the reg_read_en signal generated
   assert_next
     #(`OVL_ERROR, 1,1,0,
       `OVL_ASSERT,
       "Error! register read signal was not generated! "
       )
   u_ovl_ahb_eg_slave_reg_read
     (.clk         ( HCLK ),
      .reset_n     (HRESETn),
      .start_event ((ovl_trans_req & (~HWRITES))),
      .test_expr   (reg_read_en == 1'b1)
      );
   
   
   
   // Check register read and write operation won't assert at the same cycle
   assert_never
     #(`OVL_ERROR,
       `OVL_ASSERT,
       "Error! register read and write active at the same cycle!")
   u_ovl_ahb_eg_slave_rd_wr_illegal
     (.clk(HCLK),
      .reset_n(HRESETn),
      .test_expr((reg_write_en & reg_read_en))
      );
   
`endif
   
endmodule

