`include "vdefs.v"

module mem4
  #(
    parameter AW = 16,
    parameter MEMFILE= ""
    )
   (
    // Inputs
    input wire 		CLK,
    input wire [AW-1:2] ADDR,
    input wire [31:0] 	WDATA,
    input wire [3:0] 	WREN,
    input wire 		CS,
    
    // Outputs
    output wire [31:0] 	RDATA
    );

   localparam AWT = ((1<<(AW-2))-1);
   
   // Memory Array
   reg [7:0] 		mem4_0 [0:AWT];
   reg [7:0] 		mem4_1 [0:AWT];
   reg [7:0] 		mem4_2 [0:AWT];
   reg [7:0] 		mem4_3 [0:AWT];
   
   // Internal signals
   reg [AW-3:0] 	addr_q1;
   wire [3:0] 		write_enable;
   reg 			cs_reg;
   wire [31:0] 		read_data;

   assign write_enable[3:0] = WREN[3:0] & {4{CS}};
   
   always @ (posedge CLK)
     begin
	cs_reg <= CS;
     end
   
   // Infer Block RAM - syntax is very specific.
   always @ (posedge CLK)
     begin
	if (write_enable[0])
          mem4_0[ADDR] <= WDATA[7:0];
	if (write_enable[1])
          mem4_1[ADDR] <= WDATA[15:8];
	if (write_enable[2])
          mem4_2[ADDR] <= WDATA[23:16];
	if (write_enable[3])
          mem4_3[ADDR] <= WDATA[31:24];
	// do not use enable on read interface.
	addr_q1 <= ADDR[AW-1:2];
     end
   
   assign read_data  = {mem4_3[addr_q1],mem4_2[addr_q1],mem4_1[addr_q1],mem4_0[addr_q1]};
   
   assign RDATA      = (cs_reg) ? read_data : {32{1'b0}};
   
   integer i;
   reg [128*8:1] string;
   initial begin
      //  Initialize memory content to avoid X value on bus
      for (i = 0; i <= AWT; i=i+1)
	begin
           mem4_0[i] = 8'h00;
           mem4_1[i] = 8'h00;
           mem4_2[i] = 8'h00;
           mem4_3[i] = 8'h00;
	end
      if (MEMFILE!="")
	begin
	   string= { `SW_PATH,"/",MEMFILE,".0as" };
	   //$display("Loading memory0 from %s into %m", string);
	   $readmemh(string, mem4_0);
	   string= { `SW_PATH,"/",MEMFILE,".1as" };
	   //$display("Loading memory1 from %s into %m", string);
	   $readmemh(string, mem4_1);
	   string= { `SW_PATH,"/",MEMFILE,".2as" };
	   //$display("Loading memory2 from %s into %m", string);
	   $readmemh(string, mem4_2);
	   string= { `SW_PATH,"/",MEMFILE,".3as" };
	   //$display("Loading memory3 from %s into %m", string);
	   $readmemh(string, mem4_3);
	end
   end   

endmodule




module ahb_mem4
  #(
    parameter AW       = 12, // Address width
    parameter MEMFILE  = ""
    )
   (
    input wire 	       HCLK, // Clock
    input wire 	       HRESETn, // Reset
    // AHB inputs
    input wire 	       HSEL, // Device select
    input wire [31:0]  HADDR, // Address
    input wire [1:0]   HTRANS, // Transfer control
    input wire [2:0]   HSIZE, // Transfer size
    input wire 	       HWRITE, // Write control
    input wire [31:0]  HWDATA, // Write data
    input wire 	       HREADY, // Transfer phase done
    // AHB Outputs
    output wire        HREADYOUT, // Device ready
    output wire [31:0] HRDATA, // Read data output
    output wire        HRESP
    );  // Device response (always OKAY)

   wire [AW-3:0] 	SRAMADDR;
   wire [31:0] 		SRAMWDATA;
   wire [31:0] 		SRAMRDATA;
   wire [3:0] 		SRAMWEN;
   wire 		SRAMCS;

   // AHB to SRAM bridge
   cmsdk_ahb_to_sram #(.AW(AW)) u_ahb_to_sram
     (
      // AHB Inputs
      .HCLK       (HCLK),
      .HRESETn    (HRESETn),
      .HSEL       (HSEL),  // AHB inputs
      .HADDR      (HADDR[AW-1:0]),
      .HTRANS     (HTRANS),
      .HSIZE      (HSIZE),
      .HWRITE     (HWRITE),
      .HWDATA     (HWDATA),
      .HREADY     (HREADY),
      
      // AHB Outputs
      .HREADYOUT  (HREADYOUT), // Outputs
      .HRDATA     (HRDATA),
      .HRESP      (HRESP),
      
      // SRAM input
      .SRAMRDATA  (SRAMRDATA),
      // SRAM Outputs
      .SRAMADDR   (SRAMADDR),
      .SRAMWDATA  (SRAMWDATA),
      .SRAMWEN    (SRAMWEN),
      .SRAMCS     (SRAMCS)
      );
   
   // SRAM model
   mem4 #(.AW(AW), .MEMFILE(MEMFILE)) mem4_storage
     (
      // SRAM Inputs
      .CLK        (HCLK),
      .ADDR       (SRAMADDR),
      .WDATA      (SRAMWDATA),
      .WREN       (SRAMWEN),
      .CS         (SRAMCS),
      // SRAM Outputs
      .RDATA      (SRAMRDATA)
      );

endmodule // ahb_fpga_sram
